#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import datetime
import glob
import hashlib
import logging
import os
import shutil
import tarfile

from pycapo import CapoConfig

from . import FITSUtil as fi
from . import NGASUtil as ng
from . import archiveutils as au
from .JSONMetadataUtil import JSONMetadata
from .schema import create_session
from .schema.model import File, Filegroup, Image, ImageProduct
from .weblog_thumbs.weblogUtil import Weblog

LOG = logging.getLogger("ingestion")

session = create_session("SDM")

config = CapoConfig()


def ingestimagemetaData(imagepath, projectcode, nongas):
    """
    Ingest a image products file and a set of images
    :param imagepath: the parent directory containing the image products file and the image fits files
    :param projectcode: the project containing the images
    :param nongas: flag to indicate if ingestion to NGAS should not occur
    :return:
    """

    LOG.info("In image ingestion for project {} located in {}".format(projectcode, imagepath))
    project = au.getproject(projectcode)
    if project is None:
        LOG.error("Project {} could not be found for images located in {}".format(projectcode, imagepath))
        return False
    else:
        LOG.info("Loading images for project {} located in {}".format(project.project_code, imagepath))

    # image set filegroup --> image set
    groupname = os.path.basename(os.path.normpath(imagepath))
    imagesetfilegroup = Filegroup(project=project, groupname=groupname, type="image_set")
    session.add(imagesetfilegroup)
    session.commit()

    # image product filegroup
    image_product_filegroup = Filegroup(
        parent_filegroup_id=imagesetfilegroup.filegroup_id, groupname=groupname, type="image_products"
    )
    session.add(image_product_filegroup)

    files_to_ingest = []

    # Obtain the NGAS site/cluster info:
    site_name = config.getstring("archive-ingestion.NGASSite")
    cluster_name = config.getstring("archive-ingestion.NGASCluster")

    # Image product file
    image_product_files = glob.glob(imagepath + "/*.tar")
    image_product_file = image_product_files[0]

    if len(image_product_files) == 0:
        LOG.error("No image product files found in {}".format(imagepath))
        return False
    elif len(image_product_files) > 1:
        LOG.error("Multiple image product files found in {}".format(imagepath))
        return False
    else:
        ngas_id = ng.generate_uuids_for_ngas("evla", "imageproduct", "tar")
        product_file = File(
            file_path=imagepath,
            ngas_id=ngas_id,
            ngas_location=site_name,
            ngas_cluster=cluster_name,
            filegroup1=image_product_filegroup,
            filename=os.path.basename(image_product_file),
            filesize=os.path.getsize(image_product_file),
            format="tar",
            type="image_products",
            ingestion_time=datetime.datetime.utcnow(),
        )
        session.add(product_file)
        LOG.info(f"Moving file {image_product_file} to {os.path.join(os.path.dirname(image_product_file), ngas_id)}")
        os.link(image_product_file, os.path.join(os.path.dirname(image_product_file), ngas_id))
        files_to_ingest.append([os.path.join(imagepath, ngas_id), product_file.filesize])

    # Image product
    # Get the json metadata
    json_metadata_files = glob.glob(imagepath + "/*.json")
    if len(json_metadata_files) == 0:
        LOG.error("No json metadata file found in {}".format(imagepath))
    elif len(json_metadata_files) > 1:
        LOG.error("Multiple json metadata files found in {}".format(imagepath))
    json_metadata_file = json_metadata_files[0]
    # File doesn't exist'
    json_metadata = JSONMetadata(json_metadata_file)
    json_metadata.from_json(json_metadata_file)
    image_product = ImageProduct(
        configurations=json_metadata.get_configurations(),
        collection_name=json_metadata.get_collection_name(),
        calibration_level=json_metadata.get_calibration_level(),
        tags=json_metadata.get_product_tags(),
        product_file=product_file,
        project=project,
        image_set_filegroup=imagesetfilegroup.filegroup_id,
    )
    session.add(image_product)
    session.commit()

    # Images file group
    imagesfilegroup = Filegroup(parent_filegroup=imagesetfilegroup, groupname=groupname, type="images")
    session.add(imagesfilegroup)

    # Image file(s)
    imagefiles = glob.glob(imagepath + "/*.fits")

    # Gather the fields for the image table from the weblog - image_product_file is the .tar file
    # What to extract?

    weblog = get_weblog_from_image_products(image_product_file)

    band_codes = weblog.get_band_codes()

    for imagefile in imagefiles:
        # imagefile is the full path to a fits image
        # filename = is the file name itself
        filename = os.path.basename(imagefile)

        thumbnail_image_file = weblog.find_quicklook_thumbnail_for(filename)
        if thumbnail_image_file is not None:
            # Path to the thumbnail without and with the ROOT from CAPO
            thumbnail_path, thumbnail_destination = calculate_thumbnail_path(thumbnail_image_file)
            # Copy the thumbnail image to the thumbnails directory
            copy_thumbnail(thumbnail_image_file, thumbnail_destination)
        else:
            LOG.info(f"Skipping thumbnail for {imagefile}.")
            # Stick a NULL in the field
            thumbnail_path = None

        if "rms" in filename:
            image_type = "quicklook_image_rms"
        else:
            image_type = "quicklook_image"
        ngas_id = ng.generate_uuids_for_ngas("evla", "image", "fits")
        imgfile = File(
            file_path=imagepath,
            ngas_id=ngas_id,
            ngas_location=site_name,
            ngas_cluster=cluster_name,
            filegroup1=imagesfilegroup,
            filename=filename,
            filesize=os.path.getsize(os.path.join(imagepath, imagefile)),
            format="fits",
            type=image_type,
            ingestion_time=datetime.datetime.utcnow(),
        )
        # Gather the fields for the image table from the fits file(s)
        (
            source_name,
            telescope,
            spatial_resolution,
            field_of_view,
            min_intensity,
            max_intensity,
            ra,
            dec,
            min_freq,
            max_freq,
            rest_frequency,
            ra_element_count,
            dec_element_count,
            image_units,
            beam_axis_ratio,
            polarization_id,
            rms_noise,
            spatial_region,
            ra_pixel_size,
            dec_pixel_size,
        ) = fi.get_fits_data(imagefile)

        # Add image product id to image
        image = Image(
            target_name=source_name,
            telescope=telescope,
            thumbnail=thumbnail_path,
            spatial_resolution=spatial_resolution,
            image_field_of_view=field_of_view,
            max_intensity=max_intensity,
            min_intensity=min_intensity,
            rms_noise=rms_noise,
            polarization_id=polarization_id,
            ra=ra,
            dec=dec,
            min_frequency=min_freq,
            max_frequency=max_freq,
            ra_element_count=ra_element_count,
            dec_element_count=dec_element_count,
            starttime=json_metadata.get_starttime(),
            endtime=json_metadata.get_endtime(),
            exposure_time=json_metadata.get_endtime(),
            rest_frequency=rest_frequency,
            image_units=image_units,
            spatial_region=spatial_region,
            beam_axis_ratio=beam_axis_ratio,
            band_code=" ".join(band_codes),
            ra_pixel_size=ra_pixel_size,
            dec_pixel_size=dec_pixel_size,
            tags=json_metadata.get_image_tags(),
            file=imgfile,
            image_products_id=image_product.image_product_id,
        )

        # Image
        session.add(imgfile)
        session.add(image)

        LOG.info(f"Moving file {imagefile} to {os.path.join(os.path.dirname(imagefile), ngas_id)}")
        os.link(imagefile, os.path.join(os.path.dirname(imagefile), ngas_id))
        files_to_ingest.append([os.path.join(imagepath, ngas_id), imgfile.filesize])

    session.commit()

    # Now do the NGAS ingestion, if not disabled
    if nongas:
        LOG.info(f"NGAS ingestion is disabled. Image products and FITS images will not be ingested.")
    else:
        ingestimagefiles(files_to_ingest)


def ingestimagefiles(files_to_ingest):
    """
    Ingest fits images and image product .tar files into NGAS
    :param files_to_ingest: files to be ingested
    :return:
    """
    ng.archivefiles(files_to_ingest)


def calculate_thumbnail_path(thumbnail_image_file):
    sha1 = hashlib.sha1()

    with open(thumbnail_image_file, "rb") as f:
        for block in iter(lambda: f.read(1024), b""):
            sha1.update(block)
    key = sha1.hexdigest()

    # $1 is the first two characters of the sha1 sum, $2 the second two, and $3 the third two
    # ea887dea73301fab8b9da760014449b59e1211ae
    # $ROOT/ea/88/7d/thumbnail

    root_thumbnail_path = config.getstring("archive-ingestion.vlass.ThumbnailPath")

    # Return path without and with ROOT
    return key[0:2] + "/" + key[2:4] + "/" + key[4:6] + "/" + os.path.basename(
        thumbnail_image_file
    ), root_thumbnail_path + "/" + key[0:2] + "/" + key[2:4] + "/" + key[4:6] + "/" + os.path.basename(
        thumbnail_image_file
    )


def copy_thumbnail(thumbnail_image_file, thumbnail_destination):
    if not os.path.isdir(os.path.dirname(thumbnail_destination)):
        os.makedirs(os.path.dirname(thumbnail_destination))
    shutil.copyfile(thumbnail_image_file, thumbnail_destination)


def get_weblog_from_image_products(image_product_file):
    """
    Extract data from the weblog
    :param image_product_file: the image products tar file containing the weblog.tgz
    :return:
    """

    products_dir = os.path.dirname(image_product_file)
    product_tar_file = tarfile.open(image_product_file)

    # The pipeline gives us a uniform weblog.tgz file for VLA stuff... but ALMA
    # tacks on the sanitzed MOUS UID & the name of the processing procedures.
    files_list = product_tar_file.getnames()
    weblog_candidates = [place for place, name in enumerate(files_list) if "weblog" in name]

    # We expect a single file to match the pattern:
    if len(weblog_candidates) != 1:
        # So throw an exception when that's not the case.
        raise NameError("Cannot isolate the weblog!")

    weblog_file_name = files_list.pop(weblog_candidates[0])
    product_tar_file.extract(weblog_file_name, path=products_dir)
    # LOG.info('weblog: {}'.format(os.path.join(products_dir, weblog_file_name)))
    weblog = Weblog(products_dir, weblog_file_name)

    return weblog
